@@ -19,6 +19,7 @@
 
 #include "common.h"
 #include "colorspace.h"
+#include "hash.h"
 
 #include <libplacebo/colorspace.h>
 #include <libplacebo/tone_mapping.h>
@@ -39,6 +40,8 @@ bool pl_color_system_is_ycbcr_like(enum pl_color_system sys)
     case PL_COLOR_SYSTEM_BT_2100_HLG:
     case PL_COLOR_SYSTEM_DOLBYVISION:
     case PL_COLOR_SYSTEM_YCGCO:
+    case PL_COLOR_SYSTEM_YCGCO_RE:
+    case PL_COLOR_SYSTEM_YCGCO_RO:
         return true;
     case PL_COLOR_SYSTEM_COUNT: break;
     };
@@ -56,6 +59,8 @@ bool pl_color_system_is_linear(enum pl_color_system sys)
     case PL_COLOR_SYSTEM_SMPTE_240M:
     case PL_COLOR_SYSTEM_BT_2020_NC:
     case PL_COLOR_SYSTEM_YCGCO:
+    case PL_COLOR_SYSTEM_YCGCO_RE:
+    case PL_COLOR_SYSTEM_YCGCO_RO:
         return true;
     case PL_COLOR_SYSTEM_BT_2020_C:
     case PL_COLOR_SYSTEM_BT_2100_PQ:
@@ -69,6 +74,29 @@ bool pl_color_system_is_linear(enum pl_color_system sys)
     pl_unreachable();
 }
 
+const char *const pl_color_system_names[PL_COLOR_SYSTEM_COUNT] = {
+    [PL_COLOR_SYSTEM_UNKNOWN]       = "Auto (unknown)",
+    [PL_COLOR_SYSTEM_BT_601]        = "ITU-R Rec. BT.601 (SD)",
+    [PL_COLOR_SYSTEM_BT_709]        = "ITU-R Rec. BT.709 (HD)",
+    [PL_COLOR_SYSTEM_SMPTE_240M]    = "SMPTE-240M",
+    [PL_COLOR_SYSTEM_BT_2020_NC]    = "ITU-R Rec. BT.2020 (non-constant luminance)",
+    [PL_COLOR_SYSTEM_BT_2020_C]     = "ITU-R Rec. BT.2020 (constant luminance)",
+    [PL_COLOR_SYSTEM_BT_2100_PQ]    = "ITU-R Rec. BT.2100 ICtCp PQ variant",
+    [PL_COLOR_SYSTEM_BT_2100_HLG]   = "ITU-R Rec. BT.2100 ICtCp HLG variant",
+    [PL_COLOR_SYSTEM_DOLBYVISION]   = "Dolby Vision (invalid for output)",
+    [PL_COLOR_SYSTEM_YCGCO]         = "YCgCo (derived from RGB)",
+    [PL_COLOR_SYSTEM_YCGCO_RE]      = "YCgCo-R, even addition of bits",
+    [PL_COLOR_SYSTEM_YCGCO_RO]      = "YCgCo-R, odd addition of bits",
+    [PL_COLOR_SYSTEM_RGB]           = "Red, Green and Blue",
+    [PL_COLOR_SYSTEM_XYZ]           = "Digital Cinema Distribution Master (XYZ)",
+};
+
+const char *pl_color_system_name(enum pl_color_system sys)
+{
+    pl_assert(sys >= 0 && sys < PL_COLOR_SYSTEM_COUNT);
+    return pl_color_system_names[sys];
+}
+
 enum pl_color_system pl_color_system_guess_ycbcr(int width, int height)
 {
     if (width >= 1280 || height > 576) {
@@ -215,6 +243,33 @@ bool pl_color_primaries_is_wide_gamut(enum pl_color_primaries prim)
     pl_unreachable();
 }
 
+const char *const pl_color_primaries_names[PL_COLOR_PRIM_COUNT] = {
+    [PL_COLOR_PRIM_UNKNOWN]     = "Auto (unknown)",
+    [PL_COLOR_PRIM_BT_601_525]  = "ITU-R Rec. BT.601 (525-line = NTSC, SMPTE-C)",
+    [PL_COLOR_PRIM_BT_601_625]  = "ITU-R Rec. BT.601 (625-line = PAL, SECAM)",
+    [PL_COLOR_PRIM_BT_709]      = "ITU-R Rec. BT.709 (HD), also sRGB",
+    [PL_COLOR_PRIM_BT_470M]     = "ITU-R Rec. BT.470 M",
+    [PL_COLOR_PRIM_EBU_3213]    = "EBU Tech. 3213-E / JEDEC P22 phosphors",
+    [PL_COLOR_PRIM_BT_2020]     = "ITU-R Rec. BT.2020 (Ultra HD)",
+    [PL_COLOR_PRIM_APPLE]       = "Apple RGB",
+    [PL_COLOR_PRIM_ADOBE]       = "Adobe RGB (1998)",
+    [PL_COLOR_PRIM_PRO_PHOTO]   = "ProPhoto RGB (ROMM)",
+    [PL_COLOR_PRIM_CIE_1931]    = "CIE 1931 RGB primaries",
+    [PL_COLOR_PRIM_DCI_P3]      = "DCI-P3 (Digital Cinema)",
+    [PL_COLOR_PRIM_DISPLAY_P3]  = "DCI-P3 (Digital Cinema) with D65 white point",
+    [PL_COLOR_PRIM_V_GAMUT]     = "Panasonic V-Gamut (VARICAM)",
+    [PL_COLOR_PRIM_S_GAMUT]     = "Sony S-Gamut",
+    [PL_COLOR_PRIM_FILM_C]      = "Traditional film primaries with Illuminant C",
+    [PL_COLOR_PRIM_ACES_AP0]    = "ACES Primaries #0",
+    [PL_COLOR_PRIM_ACES_AP1]    = "ACES Primaries #1",
+};
+
+const char *pl_color_primaries_name(enum pl_color_primaries prim)
+{
+    pl_assert(prim >= 0 && prim < PL_COLOR_PRIM_COUNT);
+    return pl_color_primaries_names[prim];
+}
+
 enum pl_color_primaries pl_color_primaries_guess(int width, int height)
 {
     // HD content
@@ -234,6 +289,32 @@ enum pl_color_primaries pl_color_primaries_guess(int width, int height)
     }
 }
 
+const char *const pl_color_transfer_names[PL_COLOR_TRC_COUNT] = {
+    [PL_COLOR_TRC_UNKNOWN]      = "Auto (unknown SDR)",
+    [PL_COLOR_TRC_BT_1886]      = "ITU-R Rec. BT.1886 (CRT emulation + OOTF)",
+    [PL_COLOR_TRC_SRGB]         = "IEC 61966-2-4 sRGB (CRT emulation)",
+    [PL_COLOR_TRC_LINEAR]       = "Linear light content",
+    [PL_COLOR_TRC_GAMMA18]      = "Pure power gamma 1.8",
+    [PL_COLOR_TRC_GAMMA20]      = "Pure power gamma 2.0",
+    [PL_COLOR_TRC_GAMMA22]      = "Pure power gamma 2.2",
+    [PL_COLOR_TRC_GAMMA24]      = "Pure power gamma 2.4",
+    [PL_COLOR_TRC_GAMMA26]      = "Pure power gamma 2.6",
+    [PL_COLOR_TRC_GAMMA28]      = "Pure power gamma 2.8",
+    [PL_COLOR_TRC_PRO_PHOTO]    = "ProPhoto RGB (ROMM)",
+    [PL_COLOR_TRC_ST428]        = "Digital Cinema Distribution Master (XYZ)",
+    [PL_COLOR_TRC_PQ]           = "ITU-R BT.2100 PQ (perceptual quantizer), aka SMPTE ST2048",
+    [PL_COLOR_TRC_HLG]          = "ITU-R BT.2100 HLG (hybrid log-gamma), aka ARIB STD-B67",
+    [PL_COLOR_TRC_V_LOG]        = "Panasonic V-Log (VARICAM)",
+    [PL_COLOR_TRC_S_LOG1]       = "Sony S-Log1",
+    [PL_COLOR_TRC_S_LOG2]       = "Sony S-Log2",
+};
+
+const char *pl_color_transfer_name(enum pl_color_transfer trc)
+{
+    pl_assert(trc >= 0 && trc < PL_COLOR_TRC_COUNT);
+    return pl_color_transfer_names[trc];
+}
+
 // HLG 75% value (scene-referred)
 #define HLG_75 3.17955
 
@@ -278,12 +359,6 @@ const struct pl_hdr_metadata pl_hdr_metadata_hdr10 ={
     .max_fall = 0, // unknown
 };
 
-static const float PQ_M1 = 2610./4096 * 1./4,
-                   PQ_M2 = 2523./4096 * 128,
-                   PQ_C1 = 3424./4096,
-                   PQ_C2 = 2413./4096 * 32,
-                   PQ_C3 = 2392./4096 * 32;
-
 float pl_hdr_rescale(enum pl_hdr_scaling from, enum pl_hdr_scaling to, float x)
 {
     if (from == to)
@@ -472,6 +547,184 @@ bool pl_color_space_is_black_scaled(const struct pl_color_space *csp)
     pl_unreachable();
 }
 
+#define MAP3(...)                                                               \
+    do {                                                                        \
+        float X;                                                                \
+        for (int _i = 0; _i < 3; _i++) {                                        \
+            X = color[_i];                                                      \
+            color[_i] = __VA_ARGS__;                                            \
+        }                                                                       \
+    } while (0)
+
+void pl_color_linearize(const struct pl_color_space *csp, float color[3])
+{
+    if (csp->transfer == PL_COLOR_TRC_LINEAR)
+        return;
+
+    float csp_min, csp_max;
+    pl_color_space_nominal_luma_ex(pl_nominal_luma_params(
+        .color      = csp,
+        .metadata   = PL_HDR_METADATA_HDR10,
+        .scaling    = PL_HDR_NORM,
+        .out_min    = &csp_min,
+        .out_max    = &csp_max,
+    ));
+
+    MAP3(fmaxf(X, 0));
+
+    switch (csp->transfer) {
+    case PL_COLOR_TRC_SRGB:
+        MAP3(X > 0.04045f ? powf((X + 0.055f) / 1.055f, 2.4f)
+                          : X / 12.92f);
+        goto scale_out;
+    case PL_COLOR_TRC_BT_1886: {
+        const float lb = powf(csp_min, 1/2.4f);
+        const float lw = powf(csp_max, 1/2.4f);
+        const float a = powf(lw - lb, 2.4f);
+        const float b = lb / (lw - lb);
+        MAP3(a * powf(X + b, 2.4f));
+        return;
+    }
+    case PL_COLOR_TRC_GAMMA18: MAP3(powf(X, 1.8f)); goto scale_out;
+    case PL_COLOR_TRC_GAMMA20: MAP3(powf(X, 2.0f)); goto scale_out;
+    case PL_COLOR_TRC_UNKNOWN:
+    case PL_COLOR_TRC_GAMMA22: MAP3(powf(X, 2.2f)); goto scale_out;
+    case PL_COLOR_TRC_GAMMA24: MAP3(powf(X, 2.4f)); goto scale_out;
+    case PL_COLOR_TRC_GAMMA26: MAP3(powf(X, 2.6f)); goto scale_out;
+    case PL_COLOR_TRC_GAMMA28: MAP3(powf(X, 2.8f)); goto scale_out;
+    case PL_COLOR_TRC_PRO_PHOTO:
+        MAP3(X > 0.03125f ? powf(X, 1.8f) : X / 16);
+        goto scale_out;
+    case PL_COLOR_TRC_ST428:
+        MAP3(52.37f/48 * powf(X, 2.6f));
+        goto scale_out;
+    case PL_COLOR_TRC_PQ:
+        MAP3(powf(X, 1 / PQ_M2));
+        MAP3(fmaxf(X - PQ_C1, 0) / (PQ_C2 - PQ_C3 * X));
+        MAP3(10000 / PL_COLOR_SDR_WHITE * powf(X, 1 / PQ_M1));
+        goto scale_out;
+    case PL_COLOR_TRC_HLG: {
+        const float y = 1.2f * powf(1.111f, log2f(csp_max / HLG_REF));
+        const float b = sqrtf(3 * powf(csp_min / csp_max, 1 / y));
+        const pl_matrix3x3 rgb2xyz =
+            pl_get_rgb2xyz_matrix(pl_raw_primaries_get(csp->primaries));
+        const float *coef = rgb2xyz.m[1];
+        // OETF^-1
+        MAP3((1 - b) * X + b);
+        MAP3(X > 0.5f ? expf((X - HLG_C) / HLG_A) + HLG_B
+                      : 4 * X * X);
+        // OOTF
+        float luma = coef[0] * color[0] + coef[1] * color[1] + coef[2] * color[2];
+        luma = powf(fmaxf(luma / 12, 0), y - 1);
+        MAP3(luma * X / 12);
+        goto scale_out;
+    }
+    case PL_COLOR_TRC_V_LOG:
+        MAP3(X >= 0.181f ? powf(10, (X - VLOG_D) / VLOG_C) - VLOG_B
+                         : (X - 0.125f) / 5.6f);
+        goto scale_out;
+    case PL_COLOR_TRC_S_LOG1:
+        MAP3(powf(10, (X - SLOG_C) / SLOG_A) - SLOG_B);
+        goto scale_out;
+    case PL_COLOR_TRC_S_LOG2:
+        MAP3(X >= SLOG_Q ? (powf(10, (X - SLOG_C) / SLOG_A) - SLOG_B) / SLOG_K2
+                         : (X - SLOG_Q) / SLOG_P);
+        goto scale_out;
+    case PL_COLOR_TRC_LINEAR:
+    case PL_COLOR_TRC_COUNT:
+        break;
+    }
+
+    pl_unreachable();
+
+scale_out:
+    if (pl_color_space_is_black_scaled(csp) && csp->transfer != PL_COLOR_TRC_HLG)
+        MAP3((csp_max - csp_min) * X + csp_min);
+}
+
+void pl_color_delinearize(const struct pl_color_space *csp, float color[3])
+{
+    if (csp->transfer == PL_COLOR_TRC_LINEAR)
+        return;
+
+    float csp_min, csp_max;
+    pl_color_space_nominal_luma_ex(pl_nominal_luma_params(
+        .color      = csp,
+        .metadata   = PL_HDR_METADATA_HDR10,
+        .scaling    = PL_HDR_NORM,
+        .out_min    = &csp_min,
+        .out_max    = &csp_max,
+    ));
+
+    if (pl_color_space_is_black_scaled(csp) && csp->transfer != PL_COLOR_TRC_HLG)
+        MAP3((X - csp_min) / (csp_max - csp_min));
+
+    MAP3(fmaxf(X, 0));
+
+    switch (csp->transfer) {
+    case PL_COLOR_TRC_SRGB:
+        MAP3(X >= 0.0031308f ? 1.055f * powf(X, 1/2.4f) - 0.055f
+                             : 12.92f * X);
+        return;
+    case PL_COLOR_TRC_BT_1886: {
+        const float lb = powf(csp_min, 1/2.4f);
+        const float lw = powf(csp_max, 1/2.4f);
+        const float a = powf(lw - lb, 2.4f);
+        const float b = lb / (lw - lb);
+        MAP3(powf(X / a, 1/2.4f) - b);
+            return;
+    }
+    case PL_COLOR_TRC_GAMMA18: MAP3(powf(X, 1/1.8f)); return;
+    case PL_COLOR_TRC_GAMMA20: MAP3(powf(X, 1/2.0f)); return;
+    case PL_COLOR_TRC_UNKNOWN:
+    case PL_COLOR_TRC_GAMMA22: MAP3(powf(X, 1/2.2f)); return;
+    case PL_COLOR_TRC_GAMMA24: MAP3(powf(X, 1/2.4f)); return;
+    case PL_COLOR_TRC_GAMMA26: MAP3(powf(X, 1/2.6f)); return;
+    case PL_COLOR_TRC_GAMMA28: MAP3(powf(X, 1/2.8f)); return;
+    case PL_COLOR_TRC_ST428:
+        MAP3(powf(X * 48/52.37f, 1/2.6f));
+        return;
+    case PL_COLOR_TRC_PRO_PHOTO:
+        MAP3(X >= 0.001953f ? powf(X, 1/1.8f) : 16 * X);
+        return;
+    case PL_COLOR_TRC_PQ:
+        MAP3(powf(X * PL_COLOR_SDR_WHITE / 10000, PQ_M1));
+        MAP3(powf((PQ_C1 + PQ_C2 * X) / (1 + PQ_C3 * X), PQ_M2));
+        return;
+    case PL_COLOR_TRC_HLG: {
+        const float y = 1.2f * powf(1.111f, log2f(csp_max / HLG_REF));
+        const float b = sqrtf(3 * powf(csp_min / csp_max, 1 / y));
+        const pl_matrix3x3 rgb2xyz =
+            pl_get_rgb2xyz_matrix(pl_raw_primaries_get(csp->primaries));
+        const float *coef = rgb2xyz.m[1];
+        // OOTF^-1
+        float luma = coef[0] * color[0] + coef[1] * color[1] + coef[2] * color[2];
+        luma = fmaxf(1e-6f, powf(luma / csp_max, (1 - y) / y));
+        MAP3(12 / csp_max * luma * X);
+        // OETF
+        MAP3(X > 1 ? HLG_A * logf(X - HLG_B) + HLG_C : 0.5f * sqrtf(X));
+        MAP3((X - b) / (1 - b));
+        return;
+    }
+    case PL_COLOR_TRC_V_LOG:
+        MAP3(X >= 0.01f ? VLOG_C * log10f(X + VLOG_B) + VLOG_D
+                        : 5.6f * X + 0.125f);
+        return;
+    case PL_COLOR_TRC_S_LOG1:
+        MAP3(SLOG_A * log10f(X + SLOG_B) + SLOG_C);
+        return;
+    case PL_COLOR_TRC_S_LOG2:
+        MAP3(X >= 0 ? SLOG_A * log10f(SLOG_B * X + SLOG_C)
+                    : SLOG_P * X + SLOG_Q);
+        return;
+    case PL_COLOR_TRC_LINEAR:
+    case PL_COLOR_TRC_COUNT:
+        break;
+    }
+
+    pl_unreachable();
+}
+
 void pl_color_space_merge(struct pl_color_space *orig,
                           const struct pl_color_space *new)
 {
@@ -592,18 +845,6 @@ void pl_color_space_nominal_luma_ex(const struct pl_nominal_luma_params *params)
         *params->out_avg = avg_luma;
 }
 
-void pl_color_space_nominal_luma(const struct pl_color_space *csp,
-                                 float *out_min, float *out_max)
-{
-    pl_color_space_nominal_luma_ex(pl_nominal_luma_params(
-        .color      = csp,
-        .metadata   = PL_HDR_METADATA_ANY,
-        .scaling    = PL_HDR_NORM,
-        .out_min    = out_min,
-        .out_max    = out_max,
-    ));
-}
-
 void pl_color_space_infer(struct pl_color_space *space)
 {
     if (!space->primaries)
@@ -717,12 +958,7 @@ void pl_color_space_infer_map(struct pl_color_space *src,
 }
 
 const struct pl_color_adjustment pl_color_adjustment_neutral = {
-    .brightness     = 0.0,
-    .contrast       = 1.0,
-    .saturation     = 1.0,
-    .hue            = 0.0,
-    .gamma          = 1.0,
-    .temperature    = 0.0,
+    PL_COLOR_ADJUSTMENT_NEUTRAL
 };
 
 void pl_chroma_location_offset(enum pl_chroma_location loc, float *x, float *y)
@@ -758,9 +994,9 @@ void pl_chroma_location_offset(enum pl_chroma_location loc, float *x, float *y)
     }
 }
 
-struct pl_cie_xy pl_white_from_temp(float temp)
+struct pl_cie_xy pl_daylight_from_temp(float temp)
 {
-    temp = PL_CLAMP(temp, 2500, 25000);
+    temp = PL_CLAMP(temp, 1000, 25000);
 
     double ti = 1000.0 / temp, ti2 = ti * ti, ti3 = ti2 * ti, x;
     if (temp <= 7000) {
@@ -775,6 +1011,42 @@ struct pl_cie_xy pl_white_from_temp(float temp)
     };
 }
 
+struct pl_cie_xy pl_blackbody_from_temp(float temp)
+{
+    temp = PL_CLAMP(temp, 1667, 25000);
+
+    double ti = 1000.0 / temp, ti2 = ti * ti, ti3 = ti2 * ti, x, y;
+    if (temp <= 4000) {
+        x = -0.2661239 * ti3 - 0.2343580 * ti2 + 0.8776956 * ti + 0.179910;
+    } else {
+        x = -3.0258469 * ti3 + 2.1070379 * ti2 + 0.2226347 * ti + 0.240390;
+    }
+
+    double x2 = x * x, x3 = x2 * x;
+    if (temp <= 2222) {
+        y = -1.1063814 * x3 - 1.34811020 * x2 + 2.18555832 * x - 0.20219683;
+    } else if (temp <= 4000) {
+        y = -0.9549476 * x3 - 1.37418593 * x2 + 2.09137015 * x - 0.16748867;
+    } else {
+        y =  3.0817580 * x3 - 5.87338670 * x2 + 3.75112997 * x - 0.37001483;
+    }
+
+    return (struct pl_cie_xy) {x, y};
+}
+
+struct pl_cie_xy pl_white_from_temp(float temp)
+{
+    const struct pl_cie_xy a = pl_blackbody_from_temp(temp);
+    const struct pl_cie_xy b = pl_daylight_from_temp(temp);
+    float f = (temp - 2500) / (4000 - 2500);
+    f = PL_CLAMP(f, 0.0f, 1.0f);
+
+    return (struct pl_cie_xy) {
+        .x = PL_MIX(a.x, b.x, f),
+        .y = PL_MIX(a.y, b.y, f),
+    };
+}
+
 bool pl_raw_primaries_equal(const struct pl_raw_primaries *a,
                             const struct pl_raw_primaries *b)
 {
@@ -960,16 +1232,15 @@ pl_matrix3x3 pl_get_rgb2xyz_matrix(const struct pl_raw_primaries *prim)
     pl_matrix3x3 out = {{{0}}};
     float S[3], X[4], Z[4];
 
-    // Convert from CIE xyY to XYZ. Note that Y=1 holds true for all primaries
-    X[0] = prim->red.x   / prim->red.y;
-    X[1] = prim->green.x / prim->green.y;
-    X[2] = prim->blue.x  / prim->blue.y;
-    X[3] = prim->white.x / prim->white.y;
+    X[0] = pl_cie_X(prim->red);
+    X[1] = pl_cie_X(prim->green);
+    X[2] = pl_cie_X(prim->blue);
+    X[3] = pl_cie_X(prim->white);
 
-    Z[0] = (1 - prim->red.x   - prim->red.y)   / prim->red.y;
-    Z[1] = (1 - prim->green.x - prim->green.y) / prim->green.y;
-    Z[2] = (1 - prim->blue.x  - prim->blue.y)  / prim->blue.y;
-    Z[3] = (1 - prim->white.x - prim->white.y) / prim->white.y;
+    Z[0] = pl_cie_Z(prim->red);
+    Z[1] = pl_cie_Z(prim->green);
+    Z[2] = pl_cie_Z(prim->blue);
+    Z[3] = pl_cie_Z(prim->white);
 
     // S = XYZ^-1 * W
     for (int i = 0; i < 3; i++) {
@@ -1001,12 +1272,11 @@ pl_matrix3x3 pl_get_xyz2rgb_matrix(const struct pl_raw_primaries *prim)
     return out;
 }
 
-// LMS<-XYZ revised matrix from CIECAM97, based on a linear transform and
-// normalized for equal energy on monochrome inputs
-static const pl_matrix3x3 m_cat97 = {{
-    {  0.8562,  0.3372, -0.1934 },
-    { -0.8360,  1.8327,  0.0033 },
-    {  0.0357, -0.0469,  1.0112 },
+// Matrix used in CAT16, a revised one-step linear transform method
+static const pl_matrix3x3 m_cat16 = {{
+    {  0.401288, 0.650173, -0.051461 },
+    { -0.250268, 1.204414,  0.045854 },
+    { -0.002079, 0.048952,  0.953127 },
 }};
 
 // M := M * XYZd<-XYZs
@@ -1019,21 +1289,20 @@ static void apply_chromatic_adaptation(struct pl_cie_xy src,
     if (fabs(src.x - dest.x) < 1e-6 && fabs(src.y - dest.y) < 1e-6)
         return;
 
-    // XYZd<-XYZs = Ma^-1 * (I*[Cd/Cs]) * Ma
+    // Linear "von Kries" method, adapted from CIECAM16
     // http://www.brucelindbloom.com/index.html?Eqn_ChromAdapt.html
-    // For Ma, we use the CIECAM97 revised (linear) matrix
     float C[3][2];
 
     for (int i = 0; i < 3; i++) {
         // source cone
-        C[i][0] = m_cat97.m[i][0] * pl_cie_X(src)
-                + m_cat97.m[i][1] * 1
-                + m_cat97.m[i][2] * pl_cie_Z(src);
+        C[i][0] = m_cat16.m[i][0] * pl_cie_X(src)
+                + m_cat16.m[i][1] * 1
+                + m_cat16.m[i][2] * pl_cie_Z(src);
 
         // dest cone
-        C[i][1] = m_cat97.m[i][0] * pl_cie_X(dest)
-                + m_cat97.m[i][1] * 1
-                + m_cat97.m[i][2] * pl_cie_Z(dest);
+        C[i][1] = m_cat16.m[i][0] * pl_cie_X(dest)
+                + m_cat16.m[i][1] * 1
+                + m_cat16.m[i][2] * pl_cie_Z(dest);
     }
 
     // tmp := I * [Cd/Cs] * Ma
@@ -1041,10 +1310,10 @@ static void apply_chromatic_adaptation(struct pl_cie_xy src,
     for (int i = 0; i < 3; i++)
         tmp.m[i][i] = C[i][1] / C[i][0];
 
-    pl_matrix3x3_mul(&tmp, &m_cat97);
+    pl_matrix3x3_mul(&tmp, &m_cat16);
 
     // M := M * Ma^-1 * tmp
-    pl_matrix3x3 ma_inv = m_cat97;
+    pl_matrix3x3 ma_inv = m_cat16;
     pl_matrix3x3_invert(&ma_inv);
     pl_matrix3x3_mul(mat, &ma_inv);
     pl_matrix3x3_mul(mat, &tmp);
@@ -1126,7 +1395,7 @@ pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
                                 const struct pl_raw_primaries *prim)
 {
     // LMS<-RGB := LMS<-XYZ * XYZ<-RGB
-    pl_matrix3x3 rgb2lms = m_cat97;
+    pl_matrix3x3 rgb2lms = m_cat16;
     pl_matrix3x3 rgb2xyz = pl_get_rgb2xyz_matrix(prim);
     pl_matrix3x3_mul(&rgb2lms, &rgb2xyz);
 
@@ -1332,10 +1601,14 @@ static inline float xy_dist2(struct pl_cie_xy a, struct pl_cie_xy b)
 bool pl_primaries_compatible(const struct pl_raw_primaries *a,
                              const struct pl_raw_primaries *b)
 {
-    float RR = xy_dist2(a->red, b->red),    RG = xy_dist2(a->red, b->green),
-          RB = xy_dist2(a->red, b->blue),   GG = xy_dist2(a->green, b->green),
-          GB = xy_dist2(a->green, b->blue), BB = xy_dist2(a->blue, b->blue);
-    return RR < RG && RR < RB && GG < RG && GG < GB && BB < RB && BB < GB;
+    float RR = xy_dist2(a->red, b->red),   RG = xy_dist2(a->red, b->green),
+          RB = xy_dist2(a->red, b->blue);
+    float GR = xy_dist2(a->green, b->red), GG = xy_dist2(a->green, b->green),
+          GB = xy_dist2(a->green, b->blue);
+    float BR = xy_dist2(a->blue, b->red),  BG = xy_dist2(a->blue, b->green),
+          BB = xy_dist2(a->blue, b->blue);
+
+    return RR < RG && RR < RB && GG < GR && GG < GB && BB < BR && BB < BG;
 }
 
 // returns the intersection of the two lines defined by ab and cd
@@ -1478,6 +1751,14 @@ pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
             {1,  -1, -1},
         }};
         break;
+    case PL_COLOR_SYSTEM_YCGCO_RE:
+    case PL_COLOR_SYSTEM_YCGCO_RO:
+        m = (pl_matrix3x3) {{
+            {1, -0.5,  0.5},
+            {1,  0.5,  0  },
+            {1, -0.5, -0.5},
+        }};
+        break;
     case PL_COLOR_SYSTEM_UNKNOWN: // fall through
     case PL_COLOR_SYSTEM_RGB:
         m = pl_matrix3x3_identity;
@@ -1487,8 +1768,6 @@ pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
         // DCI-P3 primaries, which is a reasonable assumption.
         const struct pl_raw_primaries *dst = pl_raw_primaries_get(PL_COLOR_PRIM_DCI_P3);
         m = pl_get_xyz2rgb_matrix(dst);
-        // DCDM X'Y'Z' is expected to have equal energy white point (EG 432-1 Annex H)
-        apply_chromatic_adaptation((struct pl_cie_xy)CIE_E, dst->white, &m);
         break;
     }
     case PL_COLOR_SYSTEM_COUNT:
@@ -1548,6 +1827,15 @@ pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
     double ymul = 1.0 / (ymax - ymin);
     double cmul = 0.5 / (cmax - cmid);
 
+    if (repr->sys == PL_COLOR_SYSTEM_YCGCO_RE || repr->sys == PL_COLOR_SYSTEM_YCGCO_RO) {
+        int additional_bits = repr->sys == PL_COLOR_SYSTEM_YCGCO_RE ? 2 : 1;
+        double max_y = (1LL << (bit_depth - additional_bits)) - 1;
+        double max_c = (1LL << (bit_depth)) - 1;
+        ymul = cmul = max_c / max_y;
+        ymin = 0;
+        cmid = (1 << (bit_depth - 1)) / max_c;
+    }
+
     double mul[3]   = { ymul, ymul, ymul };
     double black[3] = { ymin, ymin, ymin };
 
@@ -1605,11 +1893,11 @@ bool pl_icc_profile_equal(const struct pl_icc_profile *p1,
 
 void pl_icc_profile_compute_signature(struct pl_icc_profile *profile)
 {
+    if (!profile->len)
+        profile->signature = 0;
+
     // In theory, we could get this value from the profile header itself if
     // lcms is available, but I'm not sure if it's even worth the trouble. Just
-    // hard-code this to a siphash64(), which is decently fast anyway.
-    profile->signature = pl_str_hash((pl_str) {
-        .buf = (uint8_t *) profile->data,
-        .len = profile->len
-    });
+    // hard-code this to a pl_mem_hash(), which is decently fast anyway.
+    profile->signature = pl_mem_hash(profile->data, profile->len);
 }
